"""Environment prepares a virtual environment for the PyBuild to run within.

    Every process in PyBuild requires a virtual environment to run inside to avoid any conflict
    with a cleaned system Python.

    Basic Usage:

    ```environment = Environment('pybuildenv')```

    ```with``` is available and easy to use with cleanup on ```exit```.

    ```
    with Environment('pybuildenv') as environment:
        pass
    ```
"""
import logging
import pathlib
import os
import struct
import sys

from tempfile import NamedTemporaryFile
from typing import List, Tuple, Union

from pybuild import pip
from pybuild.utils import file_utils, os_utils, process_utils

# TODO: Create snapshot=False argument where the environment
# is saved and later restored after with context completes or
# __del__ occurs.
class Environment:

    def __init__(self, env_name : str, 
                       minimum_required_python : tuple = (3, 7), 
                       maximum_required_python : tuple = (3, 9), 
                       supported_bit : tuple = (32,64)):
        """Creates an Environment for PyBuild to run in.

        Environment initialization function.

        Args:
            env_name Environment name to operate in.
        """
        assert isinstance(env_name, str)
        python_version, python_bit = self.info()
        assert isinstance(minimum_required_python, tuple) and \
               len(minimum_required_python) >= 1  and \
               python_version >= minimum_required_python, f'Python required minimum version {minimum_required_python}, received {python_version}.'
        assert isinstance(maximum_required_python, tuple) and \
               len(maximum_required_python) >= 1 and \
               python_version < maximum_required_python, f'Python required maximum version {maximum_required_python}, received {python_version}.'
        assert isinstance(supported_bit, tuple) and \
               python_bit in supported_bit, f'Python bit version must be in {supported_bit} but {python_bit} was found.'
        self.__interpreter = pathlib.Path(sys.executable)
        self.__env_name = env_name
        self.__environment_path = pathlib.Path(env_name)

        # Setup basic logging
        logging.basicConfig(
            level=logging.INFO,
            format='%(message)s',
            handlers=[
                logging.StreamHandler()
            ]
        )


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__environment_path.exists():
            os_utils.remove_directory(str(self.__environment_path))


    def __str__(self):
        if self.__environment.exists():
            return self.__environment.absolute()
        raise FileNotFoundError(f'Environment {self.__env_name} not found.')


    def __asterix_patch(self, path : pathlib.Path, prefix : str) -> Union[str, None]:
        """Returns the first file path that matches prefix.

        pathlib.Path seems to not support regular expressions in the path name when iterating over directories.
        To counteract this issue, matching the first instance and returning a path for now will be used.

        Args:
            path: Path to folder.
            prefix: Prefix to try and attempt to match.

        Returns:
            Name to first matched instance, else None.
        """
        found_path = None
        if path.exists():
            for subpath in path.iterdir():
                if subpath.name.startswith(prefix) and subpath.is_dir():
                    found_path = subpath.name
                    break
        return found_path


    def _find_interpreter(self) -> pathlib.Path:
        """Finds the Python interepter which is later used to access various executables and scripts inside the environment.

        This function would be typically called when the Python interpreter changes locations, such as one would see when a new virtual environment is stood up.
        For the redundancy of searching for new paths, the function has been protected to avoid general outside use without better reason, potentially moving the interpreter?

        Returns:
            Path reference to interpreter.
        """
        if pathlib.Path(sys.executable) == self.__interpreter:
            if os_utils.get_os() == os_utils.SupportedOS.WINDOWS:
                known_location = pathlib.Path(self.__environment_path, 'python.exe')
                if known_location.exists():
                    self.__interpreter = known_location
                else:
                    # Find the first occurrence of an interpreter - Windows only
                    for path in self.__environment_path.rglob('*'):
                        if path.is_file() and os.access(path, os.X_OK):
                            if path.with_suffix('').name == 'python' and path.suffix == '.exe':
                                self.__interpreter = path
                                break
            elif os_utils.get_os() in [os_utils.SupportedOS.LINUX, os_utils.SupportedOS.MAC]:
                known_location = pathlib.Path(self.__environment_path, 'bin', 'python')
                if known_location.exists():
                    self.__interpreter = known_location
                else:
                    raise OSError(f'Python interpreter doesn\'t exist at known location {known_location.absolute()} on Linux OS.')
        return self.__interpreter


    def cleanup(self):
        """Cleans the virtual environment if created else ignored.

        Shutil.rmtree deletes the directory, Windows this is a bit trickier and if any partition of the
        environment is currently accessed by other processes then rmtree will fail.
        """
        if self.__environment_path.exists():
            os_utils.remove_directory(str(self.__environment_path))


    def dependency_exists(self, package) -> Tuple[str, str]:
        """Checks if a dependency exists.

        Args:
            package: Package to check if it exists.

        Returns:
            Tuple of (package name, version) when the package exists else None.
        """
        pkg_info = (None, None)
        output = pip.list(self, log_output=False)
        for line in output:
            installed_package = line.rstrip().split(' ')
            if package == installed_package[0]:
                pkg_info = (installed_package[0], installed_package[-1])
        return pkg_info


    def executables(self) -> List[pathlib.Path]:
        """Returns the listing of the environments executables (Scripts: Windows, bin: Linux)

        Returns:
            List of files inside the executable directory.
        """
        if os_utils.get_os() == os_utils.SupportedOS.WINDOWS:
            return [x for x in os_utils.retrieve_directory_listing(self.__environment_path, 'Scripts') if x.is_file()]
        elif os_utils.get_os() in [os_utils.SupportedOS.LINUX, os_utils.SupportedOS.MAC]:
            return [x for x in os_utils.retrieve_directory_listing(self.__environment_path, 'bin') if x.is_file()]
        else:
            raise os_utils.PyBuildOSError()


    def info(self) -> Tuple[tuple, int]:
        """Reports environment information.

        Some build scripts may want to know the information of the environment being used,
        often this may be if the system requires a specific Python to launch the build script with.

        Returns:
            Tuple of (sys.version_info, Python Bit)
        """
        return (tuple(sys.version_info), struct.calcsize('P') * 8)


    def libs(self) -> List[pathlib.Path]:
        """Returns the listing of the environments executables (Libs: Windows, lib64, lib: Linux)

        Returns:
            List of files inside the library directory.
        """
        if os_utils.get_os() == os_utils.SupportedOS.WINDOWS:
            return os_utils.retrieve_directory_listing(self.__environment_path, 'Lib')
        elif os_utils.get_os() in [os_utils.SupportedOS.LINUX, os_utils.SupportedOS.MAC]:
            return os_utils.retrieve_directory_listing(self.__environment_path, 'lib', self.__asterix_patch(pathlib.Path(self.__environment_path, 'lib'), 'python'))
        else:
            raise os_utils.PyBuildOSError()


    def name(self) -> str:
        """Returns name of environment.
        """
        return self.__env_name


    def retrieve(self, file : str) -> List[pathlib.Path]:
        """Returns the file listing.

        Returns:
            List of files and directories inside the specified `file`.
        """
        if file in self.__environment_dict:
            return self.__environment_dict


    def python(self) -> pathlib.Path:
        """Grabs the Python interpreter located inside the environment.
        Returns:
            The known location of the Python interpreter, by default sys.executable but may be changed later.
        """
        return self.__interpreter


    def wipe(self) -> bool:
        """Wipes the entirety of the workspace of all installations.

        A developer may want the workspace they either created or are currently running in wiped prior to installing
        additional dependencies, etc. This function is only to be used by experienced users and should be avoided if possible.

        Returns:
            True if the environment was wiped successfully.

        Raises:
            OSError if the environment is unable to be wiped. The process runs through pip
        """
        successful = False
        # Open a temporaryfile
        with NamedTemporaryFile(delete=False) as tmpfd:
            tmpfd.close()
            requirements_txt = pathlib.Path(tmpfd.name)
            pip.freeze(self, tmpfd.name)
            if requirements_txt.exists():
                if file_utils.process_requirements(requirements_txt):
                    pip.uninstall(self, None, requirement=tmpfd.name)
                successful = True
            os_utils.remove_file(requirements_txt)
        return successful


import logging
import os
import subprocess

from threading import Thread

from pybuild.utils import os_utils


class _AsyncLoggingThread(Thread):
    """Async Logging Thread class"""

    def __init__(self, process, file_descriptor, logger=None, callback=None):
        """Initialization function.

        Args:
            process: Process object returned from Popen.
            file_descriptor: Process stdout or stderr.
            logger: Logging function for either info or error, should be dependent on file_descriptor passed.
            callback: Function that takes a single argument, string.
        """
        assert callable(file_descriptor.readline)
        if callback is not None:
            assert callable(callback)
        self._process = process
        self._file_descriptor = file_descriptor
        self._logger = logger
        self._callback = callback

        Thread.__init__(self)

        self._close = False


    def run(self):
        """Thread runner function to collect logging information.
        """
        while self._process.poll() is None and not self._close:
            output = self._file_descriptor.readline()
            if output:
                if self._logger is not None:
                    self._logger(str(output.rstrip(), encoding='utf-8'))
                if self._callback is not None:
                    self._callback(str(output.rstrip(), encoding='utf-8'))


    def close(self):
        self._close = True


def create_process(executable : str, command : str, wait : bool = True, log_output : bool = True, callback_func : callable = None) -> subprocess.Popen:
    """Create processes for the system to run using subprocess.

    This function will create processes that are maintained by subprocess, the subprocess will yield a return code if function blocks (waits).

    Args:
        executable: String to executable binary.
        command: Command to execute with the binary.
        wait: Wait until program has terminated.
        log_output: Log processes output to terminal, some functions may want this turned off.
        callback_func: Callback function for process to call on the event of output.

    Returns:
        On wait = True, the function will yield the return code associated with the process.
    """
    # Set Python output if calling interpreter to Unbuffered, allowing
    # print and other statements to flow through to logger
    os.environ['PYTHONUNBUFFERED'] = '1'

    async_threads = []
    # processed_command = shlex.join([executable, command])
    processed_command = ' '.join([executable, command])
    process = subprocess.Popen(processed_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=os.environ, shell=os_utils.get_os() != os_utils.SupportedOS.WINDOWS)

    for stream, log_type in [(process.stdout, logging.info if log_output else None), (process.stderr, logging.error if log_output else None)]:
        async_threads.append(_AsyncLoggingThread(process, stream, logger=log_type, callback=callback_func))
        async_threads[-1].start()

    if wait:
        process.wait()
        [x.join() for x in async_threads]
        return process.poll()
    return process


import unittest

from test_env import TestEnvironment
from test_git import TestGit
from test_pdoc import TestPdoc
from test_pip import TestPip
from test_virtualenv import TestVirtualEnv


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestEnvironment))
    suite.addTest(unittest.makeSuite(TestGit))
    suite.addTest(unittest.makeSuite(TestPdoc))
    suite.addTest(unittest.makeSuite(TestPip))
    suite.addTest(unittest.makeSuite(TestVirtualEnv))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())


import shutil
import unittest

from pathlib import Path

from pybuild import pdoc
from pybuild.environment import Environment

class TestPdoc(unittest.TestCase):

    
    def test_pdoc(self):
        with Environment('testenv') as env:
            doc_path = pdoc.make(env, 'pybuild', force=True, output_dir='test_doc')
            assert doc_path.exists(), 'Failed to create out directory for pdoc.'
            shutil.rmtree(doc_path)


if __name__ == '__main__':
    unittest.main()


""""Operating System utilities for PyBuild.

    This class acts as a utilities class which provides helpful OS related information.

"""
import logging
import pathlib
import platform
import shutil

from enum import Enum
from typing import List, Match


class PyBuildOSError(Exception):
    def __init__(self):
        super(PyBuildOSError, self).__init__(f'OS {get_os()} not supported by PyBuild.')


class SupportedOS(Enum):
    WINDOWS = 0
    LINUX = 1
    MAC = 2


def get_os() -> SupportedOS:
    """Retreive OS specifics.

    Returns:
        Enum class, SupportedOS.
    """
    operating_system = platform.system().lower()
    if operating_system == 'windows':
        return SupportedOS.WINDOWS
    elif operating_system == 'linux':
        return SupportedOS.LINUX
    elif operating_system == 'darwin':
        return SupportedOS.MAC
    else:
        raise OSError(f'Operating System {operating_system} not supported by PyBuild.')


def remove_directory(dir_name : pathlib.Path, force=True):
    """Removes the directory specified by dir_name.

    Args:
        dir_name: Name of directory to be removed.
        force: Force directory removal (Linux Only).
    """
    # TODO: rmtree has onerror which takes a call back.
    #       If the error may be traced as to which folder, permission changes and other attempts
    #       to bypass Windows silliness (.git locks on folders), removing processes from process, etc.
    try:
        shutil.rmtree(dir_name)
    except FileNotFoundError as fnfe:
        logging.error(fnfe)


def remove_file(file_name : pathlib.Path):
    """Removes the file specified by file_name.

    Args:
        file_name: Name of file to remove.
    """
    file_name.unlink()


def retrieve_directory_listing(environment : pathlib.Path, *target : str) -> List[pathlib.Path]:
    """Retreives the directory listing given the environment name provided.

    Args:
        env_name: Environment class.
        target: Single or multiple strings provided that extend from base path.

    Returns:
        List of paths from the directory listing.
    """
    if any(['..' in x for x in target]):
        raise ValueError('Backing out from directories not supported.')
    if environment.exists():
        path = pathlib.Path(environment, *target)
        if path.exists():
            return [x for x in path.iterdir()]
        raise FileNotFoundError(f'File path {path} not found.')
    raise FileNotFoundError(f'Environment {environment} not found.')



"""File system utilities for PyBuild.

    This class provides a general interface for various file related operations that a user may experience while using PyBuild.

"""
import re
import shutil

from pathlib import Path
from tempfile import NamedTemporaryFile

from pybuild.utils import os_utils


def copy(src, dst, *args):
    """Copy files and directories in PyBuild.
    
    Some builds require the copying and movement of files around the system.
    """
    return shutil.copy2(src, dst, args)


def move(src, dst):
    """Move files and directories in PyBuild.
    
    Some builds require the moving of files around the system.
    """
    return shutil.move(src, dst)


def process_requirements(requirements : Path) -> bool:
    """Processes a requirements.txt or any named variant that came off of pip.freeze.

    Certain environments have packages installed as editable which cause uninstallation related issues
    when wiping the environment using Environment.wipe. To circumvent this, processing the requirements.txt or
    any named variant that removes the editable comment and following line will fix this issue.

    Returns:
        True if the process completed successfully else False.

    Raises:
        FileNotFoundError: When requirements wasn't able to be found.
        Additional errors raised by shutil.
    """
    if requirements.exists():
        with open(requirements, 'r') as fd:
            with NamedTemporaryFile(delete=False) as tmpfd:
                index, lines = 0, fd.readlines()
                while index < len(lines):
                    line = lines[index]
                    if not '# Editable install' in line:
                        tmpfd.write(bytes(line, encoding='utf-8'))
                    else:
                        index += 1
                    index += 1
                tmpfd.flush()
                shutil.copyfile(tmpfd.name, requirements)
                tmpfd.close()
                os_utils.remove_file(Path(tmpfd.name))
                return True
    else:
        raise FileNotFoundError(f'Unable to find requirements, path provided {str(requirements)}.')


def touch(dst):
    """Creates a file.

    Creates a file on Windows, Linux or Mac dependening on the file provided in dst.s
    
    Args:
        dst: File to create.s
    """
    with open(dst, 'w') as tmp:
        pass


def validate_url(url):
    
    # URL regex validation from Django
    # https://stackoverflow.com/questions/7160737/how-to-validate-a-url-in-python-malformed-or-not
    url_regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    return re.match(url_regex, url) is not None


"""Manages documentation creation.

    Developers may want to add documentation during the build stage of their software, this can help.

    Basic Usage:

    ```
    from pybuild import pdoc
    pdoc.make('pybuild', out='doc', html=True, overwrite=True)
    ```

"""
import inspect
import logging

from pathlib import Path
from typing import List, Tuple, Union

from pybuild import pip
from pybuild.environment import Environment
from pybuild.utils import process_utils


def make(environment : Environment, 
         package : str,
         config : str=None, 
         filter : str=None,
         force : bool=False,
         html : bool=True,
         pdf : bool=False,
         output_dir : str='.',
         template_dir : str=None,
         close_stdin : str=None,
         http : Union[str,int]=None,
         skip_errors : bool=False,
         **kwargs) -> Path:
    """Creates documentation using Pdoc.

    Pdoc allows the developer to generate documentation during the build stage.
    Additional options have been provided by Pdoc that supports custom templates, etc.

    Returns:
        Path to generated documentation.

    Raises:
        ValueError: When Pdoc wasn't able to create the documentation.
        FileNotFoundError: 
    """
    arg_spec = inspect.getfullargspec(make)
    args = arg_spec.args[len(arg_spec.args) - len(arg_spec.defaults):]
    command_string = []
    for arg in args:
        processed_arg = arg.replace('_', '-')
        if str(eval(arg)) == 'True':
            command_string.append('--{}'.format(processed_arg))
        elif str(eval(arg)) not in ['False', 'None']:
            command_string.append('--{} {}'.format(processed_arg, eval(arg)))

    pip.install(environment, 'pdoc3', **kwargs)
    rc = process_utils.create_process(str(environment.python()), '-m pdoc {} {}'.format(' '.join(command_string), package))
    if rc != 0:
        raise ValueError('Failed to create documenation.')
    logging.info(f'Successfully made Pdoc documentation.')
    out_path = Path(Path.cwd(), output_dir, Path(package).name)
    if not out_path.exists():
        raise FileNotFoundError('Pdoc documentation failed to write to designated location.')
    return out_path


"""Handles everything that is pip inside PybBuild.

    This script supports installations required in PyBuild, specified either by the package itself
    or an outside developer seeking to install additional dependencies without having to communicate
    directly to Python itself, meaning, this acts as a wrapper around the process.

    Basic Usage:
    
    ```
    from pybuild import pip

    pip.install(env, 'pandas', user=False)
    ```
    
    The above will install the pandas package and the developer has requred to not install with user flag enabled.

"""
import logging

from pathlib import Path
from typing import List, Union

# from pybuild.environment import Environment
from pybuild.utils import process_utils


class Package:
    """Package class handles generics."""
    def __init__(self, package : str, version : str = None):
        """Initialization function of the class.

        Args:
            package: Package name.
            version: Version of package to install, example >=16.
        """
        self._package = package
        self._version = version


    def __str__(self):
        base_package = self._package
        if self._version:
            if '=' in self._version:
                base_package += self._version
            else:
                base_package += f'=={self._version}'
        return base_package


def freeze(environment, file_name : str) -> Path:
    """Freezes the pip dependencies of the current environment into a text file.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        file_name: File name to store the dependencies in.

    Returns:
        Path to file created through pip freeze.

    Raises:
        ValueError: Raised when process fails to execute properly.
        FileNotFoundError: Raised when pip freeze successfully runs but the file_name cannot be found.
    """
    rc = process_utils.create_process(str(environment.python()), f'-m pip freeze > {file_name}')
    if rc != 0:
        raise ValueError('Failed to freeze pip environment.')
    path = Path(Path.cwd(), file_name)
    if not path.exists():
        raise FileNotFoundError(f'Failed to find or create {file_name}.')
    return path


def install(environment, *packages : Union[Package, str], **kwargs) -> Union[Path, List[Path]]:
    """Installs packages for PyBuild environment.

    PyBuild allows the user to install packages on the fly by adding a wrapper around the pip command line.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        packages: Package to be installed, also supports multiple packages to be installed in the order provided.
        user: Enable the --user flag when calling pip from the command line. In most cases this should be True, especially on *nix systems.

    Returns:
        Return code of process launched.

    Raises:
        ValueError: Raised when process fails to execute properly.
    """
    # TODO: Process user input
    command_string = []
    for k, v in kwargs.items():
        processed_k = k.replace('_', '-')
        if str(v) == 'True':
            command_string.append('--{}'.format(processed_k))
        elif str(v) not in ['False', 'None']:
            command_string.append('--{} {}'.format(processed_k, v))
    rc = process_utils.create_process(str(environment.python()), '-m pip install {} {}'.format(' '.join(map(str, packages)) if packages else '', ' '.join(command_string)))
    if rc != 0:
        raise ValueError('Failed to install.')
    logging.info(f'Successfully installed {", ".join(map(str, packages))}.')
    return rc


def list(environment, log_output : bool = True) -> List[str]:
    """List packages for PyBuild environment.

    PyBuild allows the user to list packages from the environment by wrapping around pip through command line.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        log_output: Logs output of list to screen and log file generated.

    Returns:
        Output generated by command, may be further processed by calling function.

    Raises:
        ValueError: Raised when process fails to execute properly.
    """
    captured_output = []
    rc = process_utils.create_process(str(environment.python()), '-m pip list', log_output=log_output, callback_func=captured_output.append)
    if rc != 0:
        raise ValueError('Failed to list.')
    return captured_output


def uninstall(environment, *packages : Union[Package, str], **kwargs) -> bool:
    """Uninstall packages for PyBuild environment.

    PyBuild allows the user to uninstall packages from the environment.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        packages: Package to be uninstalled, also supports multiple packages to be uninstalled in the order provided.

    Returns:
        True if all the packages were removed successfully else False.

    Raises:
        ValueError: Raised when process fails to execute properly.
    """
    command_string = []
    for k, v in kwargs.items():
        processed_k = k.replace('_', '-')
        if str(v) == 'True':
            command_string.append('--{}'.format(processed_k))
        elif str(v) not in ['False', 'None']:
            command_string.append('--{} {}'.format(processed_k, v))

    rc = process_utils.create_process(str(environment.python()), '-m pip uninstall -y {} {}'.format(' '.join(map(str, packages)) if packages else '', ' '.join(command_string)))
    if rc != 0:
        raise ValueError('Failed to uninstall.')
    logging.info(f'Successfully uninstalled {", ".join(map(str, packages))}.')
    return rc


def upgrade(environment, *packages : Package) -> bool:
    """Upgrades package to latest.

    PyBuild allows the user to upgrade packages from the environment.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        packages: Package to be upgraded, also supports multiple packages to be installed in the order provided.

    Returns:
        True if all the packages were upgraded successfully else False.

    Raises:
        ValueError: Raised when process fails to execute properly.
    """
    rc = process_utils.create_process(str(environment.python()), '-m pip -U {}'.format(' '.join(map(str, packages)) if packages else ''))
    if rc != 0:
        raise ValueError('Failed to upgrade.')
    logging.info(f'Successfully upgraded {", ".join(map(str, packages))}.')
    return rc


import sys
import unittest

from pathlib import Path

from pybuild import pip
from pybuild.virtualenv import VirtualEnv
from pybuild.environment import Environment

class TestEnvironment(unittest.TestCase):


    def test_creation(self):
        test_env = Path('test_creation_env')
        environment = Environment(test_env.name)
        VirtualEnv(environment)
        assert test_env.exists(), 'Virtual environment wasn\'t created.'

        environment.cleanup()
        assert not test_env.exists(), 'Failed to delete virtual environment.'


    def test_with_context(self):
        test_env = Path('test_withcontext_env')
        with Environment(test_env.name) as env:
            VirtualEnv(env)
            assert test_env.exists(), 'Virtual environment wasn\'t created.'
        assert not test_env.exists(), 'Failed to delete virtual environment.'


    def test_dependency_exists(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            pip.install(env, 'matplotlib')
            dep = env.dependency_exists('matplotlib')
            assert dep[0] == 'matplotlib', 'Dependency matplotlib must exist.'
            dep = env.dependency_exists('pyinstaller')
            assert dep == (None, None), 'None must be returned.'


    def test_directories(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            executables = env.executables()
            assert len(executables) > 0, 'Executable directory must contain at least one file.'
            libs = env.libs()
            assert len(libs) > 0, 'Lib directory must contain at least one file.'


    def test_info(self):
        with Environment('testenv') as env:
            info = env.info()
            assert isinstance(info, tuple), 'Info must be of instance \'tuple\'.'
            assert isinstance(info[0], tuple), 'version_info must be of type tuple.'
            assert isinstance(info[1], int), 'Bit must be of type int.'


    def test_name(self):
        with Environment('testenv') as env:
            assert env.name() == 'testenv', 'Environment name must match instantiated name.'

    
    def test_python_func(self):
        with Environment('testenv') as env:
            assert str(env.python()) == sys.executable, 'System executable and environment executable must match.'
            VirtualEnv(env)
            assert str(env.python()) != sys.executable, 'System executable and environment executable must differ, virtual env addition.'


    def test_wipe(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            pip.install(env, 'matplotlib', 'pyinstaller', 'pandas', no_cache_dir=True)
            previous_packages = pip.list(env, log_output=False)
            env.wipe()
            current_packages = pip.list(env, log_output=False)
            assert len(previous_packages) != len(current_packages), 'Package lengths must differ after wipe.'


if __name__ == '__main__':
    unittest.main()


import shutil
import sys
import unittest

from pathlib import Path

from pybuild import pip
from pybuild import virtualenv
from pybuild.virtualenv import VirtualEnv
from pybuild.environment import Environment

class TestPip(unittest.TestCase):

    
    def test_freeze(self):
        req_file = 'test_freeze_req.txt'
        with Environment('testenv') as env:
            requirements_path = pip.freeze(env, req_file)
            assert requirements_path.exists(), f'Missing frozen file {req_file}.'
            requirements_path.unlink()


    def test_list(self):
        with Environment('testenv') as env:
            assert len(pip.list(env, log_output=False)) > 0, 'List must return at least one package.'


    def test_install(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            packages = pip.list(env, log_output=False)
            pip.install(env, 'matplotlib')
            assert len(packages) != len(pip.list(env, log_output=False)), 'Packages currently installed hasn\'t changed.'
            package, _ = env.dependency_exists('matplotlib')
            assert package == 'matplotlib', 'Missing dependency matplotlib.'


    def test_uninstall(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            pip.install(env, 'matplotlib')
            package, _ = env.dependency_exists('matplotlib')
            assert package is not None, 'Matplotlib failed to install.'
            pip.uninstall(env, 'matplotlib')
            package, _ = env.dependency_exists('matplotlib')
            assert package is None, 'Matplotlib failed to uninstall.'


if __name__ == '__main__':
    unittest.main()


"""Handles git related operations inside PyBuild.

    PyBuild allows for packages and modules not hosted on PyPi to be pulled from their respective repositories. This
    module in particular will allow for clone functionality provided by git, git will of course need
    to be installed on the system as binaries will not be stored for it to use.

    Basic Usage:

    ```
    from pybuild import git

    git.clone('https://github.com/foobar/baz.git', branch='dev', progress=True)
    ```

"""
import logging

from pathlib import Path

from pybuild.utils import file_utils, process_utils


def clone(url, destination : str=None, branch : str=None, progress : bool=True) -> Path:
    """Clones the given URL using git.

    Clone allows the developer to download additional modules to their designated locations.

    Args:
        url: URL to access git repository.
        destination: Location to store cloned repository, default is current working directory.
        branch: Branch to checkout when cloning, leave None for default.
        progress: Display progress bar while cloning repository.

    Returns:
        Path to cloned repository, destination will be this return value.
        None if the function fails to find validate the repository successfully.
        
    Raises:
        ValueError: Raised when process fails to execute properly.
        FileNotFoundError: Raised when git clone successfully runs but the destination cannot be found.

    """
    if file_utils.validate_url(url):
        url_path = Path(url)
        destination = destination if destination else url_path.with_suffix('').name
        git_command = f'clone {url} {destination}'
        git_command += f' -b {branch}' if branch else ''
        git_command += ' --progress' if progress else ''
        rc = process_utils.create_process('git', git_command)
        if rc != 0:
            raise ValueError('Failed to clone repository.')
        cloned_path = Path(Path.cwd(), destination)
        if not cloned_path.exists():
            raise FileNotFoundError(f'Failed to find {destination}.')
        return cloned_path
    else:
        logging.error('Invalid URL detected.')



"""Virtual Environment class that handles installation designated by the user.

    A user may want to install additional dependencies to their base system Python or the Python the ran PyBuild with, effectively being
    regarded as the systems interpreter. In the event of this, VirtualEnv was stood up to allow the user freedom of creating a virtual environment
    whenever and allowing themselves to install dependencies using pip prior to the virtual environments creation.

    Environment has a hard set limitation based on _find_interpreter that only runs once, after the sys.executable and __interpreter differ
    the function won't operate again. To avoid this, create multiple environments where you need additional virtual environments.

    Basic Usage:
    with Environment('pybuildenv') as environment:
        virtual_environment_pkg = VirtualEnv(environment)
        ...

        second_environment = Environment('testenv')
        virtual_environment_pkg = VirtualEnv(second_environment)
        ...
        second_environment.cleanup()

    The return value of VirtualEnv is typically useless. Packages currently are only used for displaying the package and version in a way for pip to understand.

"""
from pybuild import pip
from pybuild.environment import Environment
from pybuild.utils import process_utils


class VirtualEnv(pip.Package):
    """Virtual Environment Class."""
    def __init__(self, environment : Environment, version : str = None, user : bool = False):
        super().__init__('virtualenv', version=version)
        
        if environment.dependency_exists('virtualenv') == (None, None):
            pip.install(environment, self, user=user)
        if process_utils.create_process(str(environment.python()), f'-m virtualenv {environment.name()}') != 0:
            raise OSError('Failed to create virtual environment.')
        environment._find_interpreter()




import unittest

from pathlib import Path

from pybuild.virtualenv import VirtualEnv
from pybuild.environment import Environment

class TestVirtualEnv(unittest.TestCase):

    
    def test_virtualenv(self):
        with Environment('testenv') as env:
            VirtualEnv(env)
            assert Path('testenv').exists(), 'Failed to find virtual environment folder.'
            

if __name__ == '__main__':
    unittest.main()


"""Creates Python executable inside the provided environment.

    PyBuild supports the developer in building an executable
    of either their entire program or specific modules for said program,
    configurability is everything.

    Basic Usage:

"""
import logging

from pybuild import pip
from pybuild.environment import Environment
from pybuild.utils import process_utils


def make_executable(environment : Environment, command : str) -> int:
    """Creates the executable using PyInstaller.

    PyInstaller creates executable binaries for Windows, Linux, and macOS. The arguments provided must match what's passed 
    to PyInstaller command, later this function will be refined.

    Args:
        environment: Environment where PyBuild exists, this was initialized during the Environment initialization stage.
        command: Command to pass to PyInstaller.
    
    Returns:
        Return code of the subprocess executed with PyInstaller.
    """
    if environment.dependency_exists('pyinstaller') is None:
        pip.install(environment, 'pyinstaller')
    pyinstaller_executable = [x for x in environment.executables() if 'pyinstaller' in x.name][0]
    return process_utils.create_process(str(pyinstaller_executable), command)


from setuptools import setup, find_packages

__version__ = "0.2.0"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name = 'pybuildme',
    python_requires='>=3.7',
    packages = find_packages(),
    install_requires=[],
    version = __version__,
    description = 'PyBuild is a build environment for Python that starts from the virtual environment upwards, must be installed on base system prior to running.',
    # long_description = long_description,
    author = 'Kyle Darling',
    author_email = 'kdarling95@yahoo.com',
    url = 'https://github.com/IAmAbszol/PyBuild',
    download_url = '',
    keywords = ['build', 'python', 'automation', 'git', 'framework'],
    classifiers = [],
    license = "MIT",
    include_package_data=True
)



import unittest

from pathlib import Path

from pybuild.virtualenv import VirtualEnv
from pybuild.environment import Environment

class TestGit(unittest.TestCase):

    
    def test_git(self):
        """
        TODO:   Add clone test here. When deleting a cloned repository, .git causes rmtree to fail.
                Recursively going through the contents of .git, bottom up, deleting the files might resolve this.
        """
        pass

if __name__ == '__main__':
    unittest.main()